% MPC computation for a simple 2D system (unstable).
% Requires YALMIP and mosek installation
% ME 234(b), Spring 2022
% Anushri Dixit, Joel Burdick
%%
yalmip('clear'); clear all; close all;
nx = 2; nu = 1;
A = [1 0;0 1.1]; 
B = [1; 0.5];

% Optimal Control Parameters
Q1 = 1.*eye(nx);
Q2 = 100.*[3 -2; -2 4];
R = 1.*eye(nu);
Q=Q1;
N1 = 2;
N2 = 12;
N3 = 50;
N4 = 1000;
N= N1;
P = dare(A, B, Q, R);


% Initial condition (IC)
IC = [1;1];
% Written as a decision variable so that we can solve for an arbitrary IC
x0 = sdpvar(nx,1);           

% Control Input as the decision variable
u = sdpvar(repmat(nu,1,N),repmat(1,1,N));

% Set up constraints
constraints = [];
objective = 0;
x = x0;
for k = 1:N
 x = A*x + B*u{k};
 objective = objective + x'*Q*x + u{k}'*R*u{k};
 constraints = [constraints, -1 <= u{k}<= 1, [-5;-5]<=x<=[5;5]];
end

% Terminal cost P_\infty
objective = objective + x'*P*x;

% Terminal constraint X_f = {0}
% constraints = [constraints, x == [0;0]];

% Terminal constraint X_f = control invariant set obtained from MPT3
% example
load('invariant_set.mat')
constraints = [constraints, invariant*[x;-1] <= 0];

% Find controller as a function of the IC
ops = sdpsettings('solver', 'mosek', 'verbose',1);
controller = optimizer(constraints,objective,ops,x0,u{1});

appliedU = [];
x_ic = IC;
clf;
X = x_ic;

hold on
for i = 1:200
  display(['iteration: ', num2str(i)])
  U = controller{x_ic};
  x_ic = A*x_ic + B*U(:,1);
  X = [X x_ic];
  appliedU = [appliedU; U(1)];
end
plot(X(1,:), X(2,:), '*-b')
xlabel('x_1')
ylabel('x_2')